@@ -23,6 +23,7 @@
 #
 import functools
 import io
+import operator
 import os
 import struct
 from binascii import unhexlify
@@ -30,7 +31,7 @@ from functools import reduce
 from io import BytesIO
 from operator import and_, or_
 from struct import pack, unpack
-from typing import Any, BinaryIO, Dict, List, Optional, Tuple
+from typing import Any, BinaryIO, Optional, Union
 
 from py7zr.compressor import SevenZipCompressor, SevenZipDecompressor
 from py7zr.exceptions import Bad7zFile
@@ -42,12 +43,27 @@ P7ZIP_MAJOR_VERSION = b"\x00"
 P7ZIP_MINOR_VERSION = b"\x04"
 
 
-def read_crcs(file: BinaryIO, count: int) -> List[int]:
+class WriteWithCrc(io.RawIOBase):
+    """Thin wrapper for file object to calculate crc32 when write called."""
+
+    def __init__(self, fp: BinaryIO):
+        self._fp = fp
+        self.digest = 0
+
+    def write(self, data):
+        self.digest = calculate_crc32(data, self.digest)
+        return self._fp.write(data)
+
+    def tell(self):
+        return self._fp.tell()
+
+
+def read_crcs(file: BinaryIO, count: int) -> list[int]:
     data = file.read(4 * count)
     return [unpack("<L", data[i * 4 : i * 4 + 4])[0] for i in range(count)]
 
 
-def write_crcs(file: BinaryIO, crcs):
+def write_crcs(file: Union[BinaryIO, WriteWithCrc], crcs):
     for crc in crcs:
         write_uint32(file, crc)
 
@@ -56,30 +72,30 @@ def read_byte(file: BinaryIO) -> int:
     return ord(file.read(1))
 
 
-def write_bytes(file: BinaryIO, data: bytes):
+def write_bytes(file: Union[BinaryIO, WriteWithCrc], data: bytes):
     return file.write(data)
 
 
-def write_byte(file: BinaryIO, data):
+def write_byte(file: Union[BinaryIO, WriteWithCrc], data):
     assert len(data) == 1
     return write_bytes(file, data)
 
 
-def read_real_uint64(file: BinaryIO) -> Tuple[int, bytes]:
+def read_real_uint64(file: BinaryIO) -> tuple[int, bytes]:
     """read 8 bytes, return unpacked value as a little endian unsigned long long, and raw data."""
     res = file.read(8)
     a = unpack("<Q", res)[0]
     return a, res
 
 
-def read_uint32(file: BinaryIO) -> Tuple[int, bytes]:
+def read_uint32(file: BinaryIO) -> tuple[int, bytes]:
     """read 4 bytes, return unpacked value as a little endian unsigned long, and raw data."""
     res = file.read(4)
     a = unpack("<L", res)[0]
     return a, res
 
 
-def write_uint32(file: BinaryIO, value):
+def write_uint32(file: Union[BinaryIO, WriteWithCrc], value):
     """write uint32 value in 4 bytes."""
     b = pack("<L", value)
     file.write(b)
@@ -115,12 +131,12 @@ def read_uint64(file: BinaryIO) -> int:
     return value + (highpart << (vlen * 8))
 
 
-def write_real_uint64(file: BinaryIO, value: int):
+def write_real_uint64(file: Union[BinaryIO, WriteWithCrc], value: int):
     """write 8 bytes, as an unsigned long long."""
     file.write(pack("<Q", value))
 
 
-def write_uint64(file: BinaryIO, value: int):
+def write_uint64(file: Union[BinaryIO, WriteWithCrc], value: int):
     """
     UINT64 means real UINT64 encoded with the following scheme:
 
@@ -138,7 +154,7 @@ def write_uint64(file: BinaryIO, value: int):
     if value < 0x80:
         file.write(pack("B", value))
         return
-    if value > 0x01FFFFFFFFFFFFFF:
+    if value > 0xFFFFFFFFFFFFFF:
         file.write(b"\xff")
         file.write(value.to_bytes(8, "little"))
         return
@@ -158,7 +174,7 @@ def write_uint64(file: BinaryIO, value: int):
         file.write(ba)
 
 
-def read_boolean(file: BinaryIO, count: int, checkall: bool = False) -> List[bool]:
+def read_boolean(file: BinaryIO, count: int, checkall: bool = False) -> list[bool]:
     if checkall:
         all_defined = file.read(1)
         if all_defined != unhexlify("00"):
@@ -175,7 +191,7 @@ def read_boolean(file: BinaryIO, count: int, checkall: bool = False) -> List[boo
     return result
 
 
-def write_boolean(file: BinaryIO, booleans: List[bool], all_defined: bool = False):
+def write_boolean(file: Union[BinaryIO, WriteWithCrc], booleans: list[bool], all_defined: bool = False):
     if all_defined and reduce(and_, booleans, True):
         file.write(b"\x01")
         return
@@ -190,16 +206,16 @@ def write_boolean(file: BinaryIO, booleans: List[bool], all_defined: bool = Fals
 
 def read_utf16(file: BinaryIO) -> str:
     """read a utf-16 string from file"""
-    val = ""
+    val = b""
     for _ in range(MAX_LENGTH):
         ch = file.read(2)
         if ch == unhexlify("0000"):
             break
-        val += ch.decode("utf-16LE")
-    return val
+        val += ch
+    return val.decode("utf-16LE")
 
 
-def write_utf16(file: BinaryIO, val: str):
+def write_utf16(file: Union[BinaryIO, WriteWithCrc], val: str):
     """write a utf-16 string to file"""
     for c in val:
         file.write(c.encode("utf-16LE"))
@@ -226,9 +242,9 @@ class PackInfo:
     def __init__(self) -> None:
         self.packpos: int = 0
         self.numstreams: int = 0
-        self.packsizes: List[int] = []
-        self.digestdefined: List[bool] = []
-        self.crcs: List[int] = []
+        self.packsizes: list[int] = []
+        self.digestdefined: list[bool] = []
+        self.crcs: list[int] = []
         self.enable_digests: bool = True
 
     @classmethod
@@ -243,18 +259,19 @@ class PackInfo:
             self.packsizes = [read_uint64(file) for _ in range(self.numstreams)]
             pid = file.read(1)
             if pid == PROPERTY.CRC:
-                self.enable_digests = True
                 self.digestdefined = read_boolean(file, self.numstreams, True)
                 for crcexist in self.digestdefined:
+                    self.enable_digests = True
                     if crcexist:
                         self.crcs.append(read_uint32(file)[0])
                 pid = file.read(1)
         if pid != PROPERTY.END:
             raise Bad7zFile("end id expected but %s found" % repr(pid))  # pragma: no-cover  # noqa
-        self.packpositions = [sum(self.packsizes[:i]) for i in range(self.numstreams + 1)]  # type: List[int]
+        self.packpositions: list[int] = [sum(self.packsizes[:i]) for i in range(self.numstreams + 1)]
+        self.enable_digests = len(self.crcs) > 0
         return self
 
-    def write(self, file: BinaryIO):
+    def write(self, file: Union[BinaryIO, WriteWithCrc]):
         assert self.packpos is not None
         assert self.numstreams == len(self.packsizes)
         write_byte(file, PROPERTY.PACK_INFO)
@@ -263,6 +280,7 @@ class PackInfo:
         write_byte(file, PROPERTY.SIZE)
         for size in self.packsizes:
             write_uint64(file, size)
+        self.enable_digests = functools.reduce(operator.or_, self.digestdefined, self.enable_digests)
         if self.enable_digests:
             assert len(self.crcs) == self.numstreams
             write_byte(file, PROPERTY.CRC)
@@ -311,10 +329,10 @@ class Folder:
     ]
 
     def __init__(self) -> None:
-        self.unpacksizes: List[int] = []
-        self.coders: List[Dict[str, Any]] = []
-        self.bindpairs: List[Bond] = []
-        self.packed_indices: List[int] = []
+        self.unpacksizes: list[int] = []
+        self.coders: list[dict[str, Any]] = []
+        self.bindpairs: list[Bond] = []
+        self.packed_indices: list[int] = []
         # calculated values
         # internal values
         self.solid: bool = False
@@ -343,7 +361,7 @@ class Folder:
             iscomplex = b & 0x10 == 0x10
             hasattributes = b & 0x20 == 0x20
             if methodsize > 0:
-                c: Dict[str, Any] = {"method": file.read(methodsize)}
+                c: dict[str, Any] = {"method": file.read(methodsize)}
             else:
                 c = {"method": b"\x00"}
             if iscomplex:
@@ -385,7 +403,7 @@ class Folder:
         self.digestdefined = False
         num_bindpairs = sum([c["numoutstreams"] for c in self.coders]) - 1
         self.bindpairs = [Bond(incoder=i + 1, outcoder=i) for i in range(num_bindpairs)]
-        # Only simple codecs are suport, assert it
+        # Only simple codecs are support, assert it
         assert sum([c["numinstreams"] for c in self.coders]) == sum([c["numoutstreams"] for c in self.coders])
 
     def write(self, file: BinaryIO):
@@ -468,7 +486,7 @@ class UnpackInfo:
     def _read(self, file: BinaryIO):
         pid = file.read(1)
         if pid != PROPERTY.FOLDER:
-            raise Bad7zFile("folder id expected but %s found" % repr(pid))  # pragma: no-cover
+            raise Bad7zFile(f"folder id expected but {repr(pid)} found")  # pragma: no-cover
         self.numfolders = read_uint64(file)
         self.folders = []
         external = read_byte(file)
@@ -485,7 +503,7 @@ class UnpackInfo:
     def _retrieve_coders_info(self, file: BinaryIO):
         pid = file.read(1)
         if pid != PROPERTY.CODERS_UNPACK_SIZE:
-            raise Bad7zFile("coders unpack size id expected but %s found" % repr(pid))  # pragma: no-cover
+            raise Bad7zFile(f"coders unpack size id expected but {repr(pid)} found")  # pragma: no-cover
         for folder in self.folders:
             for c in folder.coders:
                 for _ in range(c["numoutstreams"]):
@@ -499,11 +517,9 @@ class UnpackInfo:
                 folder.crc = crcs[idx]
             pid = file.read(1)
         if pid != PROPERTY.END:
-            raise Bad7zFile(
-                "end id expected but 0x{:02x} found at 0x{:08x}".format(ord(pid), file.tell())
-            )  # pragma: no-cover  # noqa
+            raise Bad7zFile(f"end id expected but 0x{ord(pid):02x} found at 0x{file.tell():08x}")  # pragma: no-cover  # noqa
 
-    def write(self, file: BinaryIO):
+    def write(self, file: Union[BinaryIO, WriteWithCrc]):
         assert self.numfolders == len(self.folders)
         file.write(PROPERTY.UNPACK_INFO)
         file.write(PROPERTY.FOLDER)
@@ -536,18 +552,18 @@ class SubstreamsInfo:
     ]
 
     def __init__(self):
-        self.digests: List[int] = []
-        self.digestsdefined: List[bool] = []
-        self.unpacksizes: Optional[List[int]] = None
-        self.num_unpackstreams_folders: List[int] = []
+        self.digests: list[int] = []
+        self.digestsdefined: list[bool] = []
+        self.unpacksizes: Optional[list[int]] = None
+        self.num_unpackstreams_folders: list[int] = []
 
     @classmethod
-    def retrieve(cls, file: BinaryIO, numfolders: int, folders: List[Folder]):
+    def retrieve(cls, file: BinaryIO, numfolders: int, folders: list[Folder]):
         obj = cls()
         obj._read(file, numfolders, folders)
         return obj
 
-    def _read(self, file: BinaryIO, numfolders: int, folders: List[Folder]):
+    def _read(self, file: BinaryIO, numfolders: int, folders: list[Folder]):
         pid = file.read(1)
         if pid == PROPERTY.NUM_UNPACK_STREAM:
             self.num_unpackstreams_folders = [read_uint64(file) for _ in range(numfolders)]
@@ -588,12 +604,12 @@ class SubstreamsInfo:
                         didx += 1
             pid = file.read(1)
         if pid != PROPERTY.END:
-            raise Bad7zFile("end id expected but %r found" % pid)  # pragma: no-cover
+            raise Bad7zFile(f"end id expected but {repr(pid)} found")  # pragma: no-cover
         if not self.digestsdefined:
             self.digestsdefined = [False] * num_digests_total
             self.digests = [0] * num_digests_total
 
-    def write(self, file: BinaryIO):
+    def write(self, file: Union[BinaryIO, WriteWithCrc]):
         if len(self.num_unpackstreams_folders) == 0:  # pragma: no-cover  # nothing to write
             return
         write_byte(file, PROPERTY.SUBSTREAMS_INFO)
@@ -649,9 +665,9 @@ class StreamsInfo:
             self.substreamsinfo = SubstreamsInfo.retrieve(file, self.unpackinfo.numfolders, self.unpackinfo.folders)
             pid = file.read(1)
         if pid != PROPERTY.END:
-            raise Bad7zFile("end id expected but %s found" % repr(pid))  # pragma: no-cover
+            raise Bad7zFile(f"end id expected but {repr(pid)} found")  # pragma: no-cover
 
-    def write(self, file: BinaryIO):
+    def write(self, file: Union[BinaryIO, WriteWithCrc]):
         write_byte(file, PROPERTY.MAIN_STREAMS_INFO)
         if self.packinfo is not None:
             self.packinfo.write(file)
@@ -673,7 +689,7 @@ class HeaderStreamsInfo(StreamsInfo):
         self.unpackinfo: UnpackInfo = UnpackInfo()
         self.unpackinfo.numfolders = 1
 
-    def write(self, file: BinaryIO):
+    def write(self, file: Union[BinaryIO, WriteWithCrc]):
         write_byte(file, PROPERTY.ENCODED_HEADER)
         self.packinfo.write(file)
         self.unpackinfo.write(file)
@@ -686,8 +702,8 @@ class FilesInfo:
     __slots__ = ["files", "emptyfiles", "antifiles"]
 
     def __init__(self):
-        self.files: List[Dict[str, Any]] = []
-        self.emptyfiles: List[bool] = []
+        self.files: list[dict[str, Any]] = []
+        self.emptyfiles: list[bool] = []
         self.antifiles = None
 
     @classmethod
@@ -747,13 +763,13 @@ class FilesInfo:
             elif prop == PROPERTY.START_POS:
                 self._read_start_pos(buffer)
             else:
-                raise Bad7zFile("invalid type %r" % prop)  # pragma: no-cover
+                raise Bad7zFile(f"invalid type {repr(prop)}")  # pragma: no-cover
 
     def _read_name(self, buffer: BinaryIO) -> None:
         for f in self.files:
             f["filename"] = read_utf16(buffer).replace("\\", "/")
 
-    def _read_attributes(self, buffer: BinaryIO, defined: List[bool]) -> None:
+    def _read_attributes(self, buffer: BinaryIO, defined: list[bool]) -> None:
         for idx, f in enumerate(self.files):
             f["attributes"] = read_uint32(buffer)[0] if defined[idx] else None
 
@@ -773,15 +789,17 @@ class FilesInfo:
         for i, f in enumerate(self.files):
             f["startpos"] = read_real_uint64(fp)[0] if defined[i] else None
 
-    def _write_times(self, fp: BinaryIO, propid, name: str) -> None:
+    def _write_times(self, fp: Union[BinaryIO, WriteWithCrc], propid, name: str) -> None:
         write_byte(fp, propid)
-        defined = []  # type: List[bool]
-        num_defined = 0  # type: int
+        defined: list[bool] = []
+        num_defined = 0
         for f in self.files:
             if name in f.keys():
                 if f[name] is not None:
                     defined.append(True)
                     num_defined += 1
+                    continue
+            defined.append(False)
         size = num_defined * 8 + 2
         if not reduce(and_, defined, True):
             size += bits_to_bytes(num_defined)
@@ -794,7 +812,7 @@ class FilesInfo:
             else:
                 pass
 
-    def _write_prop_bool_vector(self, fp: BinaryIO, propid, vector) -> None:
+    def _write_prop_bool_vector(self, fp: Union[BinaryIO, WriteWithCrc], propid, vector) -> None:
         write_byte(fp, propid)
         write_boolean(fp, vector, all_defined=False)
 
@@ -805,7 +823,7 @@ class FilesInfo:
                 return True
         return False
 
-    def _write_names(self, file: BinaryIO):
+    def _write_names(self, file: Union[BinaryIO, WriteWithCrc]):
         name_defined = 0
         names = []
         name_size = 0
@@ -822,7 +840,7 @@ class FilesInfo:
                 write_utf16(file, n)
 
     def _write_attributes(self, file):
-        defined = []  # type: List[bool]
+        defined: list[bool] = []
         num_defined = 0
         for f in self.files:
             if "attributes" in f.keys() and f["attributes"] is not None:
@@ -841,7 +859,7 @@ class FilesInfo:
             if defined[i]:
                 write_uint32(file, f["attributes"])
 
-    def write(self, file: BinaryIO):
+    def write(self, file: Union[BinaryIO, WriteWithCrc]):
         assert self.files is not None
         write_byte(file, PROPERTY.FILES_INFO)
         numfiles = len(self.files)
@@ -877,21 +895,6 @@ class FilesInfo:
         write_byte(file, PROPERTY.END)
 
 
-class WriteWithCrc:
-    """Thin wrapper for file object to calculate crc32 when write called."""
-
-    def __init__(self, fp: BinaryIO):
-        self._fp = fp
-        self.digest = 0
-
-    def write(self, data):
-        self.digest = calculate_crc32(data, self.digest)
-        return self._fp.write(data)
-
-    def tell(self):
-        return self._fp.tell()
-
-
 class Header:
     """the archive header"""
 
@@ -899,12 +902,13 @@ class Header:
 
     def __init__(self) -> None:
         self.solid: bool = False
-        self.main_streams = None
-        self.files_info = None
+        self.main_streams: Optional[StreamsInfo] = None
+        self.files_info: Optional[FilesInfo] = None
         self.size: int = 0
         self._start_pos: int = 0
         self.password: Optional[str] = None
         self._initialized: bool = False
+        self.filters: Optional[list[dict[str, int]]] = None
 
     @classmethod
     def retrieve(cls, fp: BinaryIO, buffer: BytesIO, start_pos: int, password=None):
@@ -928,7 +932,7 @@ class Header:
             self._extract_header_info(buffer)
             return
         if pid != PROPERTY.ENCODED_HEADER:
-            raise TypeError("Unknown field: %r" % id)  # pragma: no-cover
+            raise TypeError(f"Unknown field: {repr(pid)}")  # pragma: no-cover
         # get from encoded header
         streams = HeaderStreamsInfo.retrieve(buffer)
         buffer2 = io.BytesIO()
@@ -951,13 +955,14 @@ class Header:
             self.size += compressed_size
             src_start += compressed_size
             if folder.digestdefined:
+                streams.packinfo.enable_digests = True
                 if folder.crc != calculate_crc32(folder_data):
                     raise Bad7zFile("invalid block data")
             buffer2.write(folder_data)
         buffer2.seek(0, 0)
         pid = buffer2.read(1)
         if pid != PROPERTY.HEADER:
-            raise TypeError("Unknown field: %r" % pid)  # pragma: no-cover
+            raise TypeError(f"Unknown field: {repr(pid)}")  # pragma: no-cover
         self._extract_header_info(buffer2)
 
     def _encode_header(self, file: BinaryIO, afterheader: int, filters):
@@ -986,7 +991,7 @@ class Header:
         # actual header start position
         startpos = file.tell()
         crcfile = WriteWithCrc(file)
-        headerstreams.write(crcfile)  # type: ignore  # noqa
+        headerstreams.write(crcfile)
         digest = crcfile.digest
         return startpos, digest
 
@@ -1000,12 +1005,12 @@ class Header:
             startpos, headercrc = self._encode_header(file, afterheader, filters)
         else:
             crcfile = WriteWithCrc(file)
-            write_byte(crcfile, PROPERTY.HEADER)  # type: ignore  # noqa
+            write_byte(crcfile, PROPERTY.HEADER)
             if self.main_streams is not None:
                 self.main_streams.write(crcfile)
             if self.files_info is not None:
                 self.files_info.write(crcfile)
-            write_byte(crcfile, PROPERTY.END)  # type: ignore
+            write_byte(crcfile, PROPERTY.END)
             headercrc = crcfile.digest
         endpos = file.tell()
         header_len = endpos - startpos
@@ -1020,7 +1025,7 @@ class Header:
             self.files_info = FilesInfo.retrieve(fp)
             pid = fp.read(1)
         if pid != PROPERTY.END:
-            raise Bad7zFile("end id expected but %s found" % (repr(pid)))  # pragma: no-cover
+            raise Bad7zFile(f"end id expected but {repr(pid)} found")  # pragma: no-cover
 
     @staticmethod
     def build_header(filters, password):
@@ -1032,20 +1037,22 @@ class Header:
     def initialize(self):
         if not self._initialized:
             self._initialized = True
+            folder = Folder()
+            folder.password = self.password
+            folder.prepare_coderinfo(self.filters)
             if self.main_streams is not None:
                 # append mode
-                folder = Folder()
-                folder.password = self.password
-                folder.prepare_coderinfo(self.filters)
-                self.main_streams.packinfo.enable_digests = False  # FIXME
-                self.main_streams.unpackinfo.folders.append(folder)
-                self.main_streams.unpackinfo.numfolders += 1
-                self.main_streams.substreamsinfo.num_unpackstreams_folders.append(0)
+                if self.main_streams.unpackinfo is not None:
+                    self.main_streams.unpackinfo.folders.append(folder)
+                    self.main_streams.unpackinfo.numfolders += 1
+                else:
+                    pass  # unexpected
+                if self.main_streams.substreamsinfo is not None:
+                    self.main_streams.substreamsinfo.num_unpackstreams_folders.append(0)
+                else:
+                    pass  # unexpected
             else:
                 # create new header
-                folder = Folder()
-                folder.password = self.password
-                folder.prepare_coderinfo(self.filters)
                 folders = [folder]
                 self.files_info = FilesInfo()
                 self.main_streams = StreamsInfo()
@@ -1065,6 +1072,11 @@ class Header:
                 self.main_streams.substreamsinfo.num_unpackstreams_folders = [0]
                 self.main_streams.packinfo.packsizes = []
                 self.main_streams.packinfo.crcs = []
+            return folder
+        else:
+            if self.main_streams is not None and self.main_streams.unpackinfo is not None:
+                return self.main_streams.unpackinfo.folders[-1]
+        return None  # unexpected
 
 
 class SignatureHeader:
@@ -1079,10 +1091,10 @@ class SignatureHeader:
     ]
 
     def __init__(self) -> None:
-        self.version = (
+        self.version: tuple[bytes, ...] = (
             P7ZIP_MAJOR_VERSION,
             P7ZIP_MINOR_VERSION,
-        )  # type: Tuple[bytes, ...]
+        )
         self.startheadercrc: int = -1
         self.nextheaderofs: int = -1
         self.nextheadersize: int = -1
@@ -1133,7 +1145,7 @@ class SignatureHeader:
         write_real_uint64(file, self.nextheadersize)
         write_uint32(file, self.nextheadercrc)
 
-    def _write_skelton(self, file: BinaryIO):
+    def _write_skeleton(self, file: BinaryIO):
         file.seek(0, 0)
         write_bytes(file, MAGIC_7Z)
         write_byte(file, self.version[0])
